#define PY_SSIZE_T_CLEAN
#include <Python.h>
#include <structmember.h>

typedef struct {
  PyObject_HEAD
  unsigned long t_min;
  unsigned long t_max;
} cvs2svn_timerange;

static PyObject *
cvs2svn_timerange_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  cvs2svn_timerange *self = (cvs2svn_timerange *)type->tp_alloc(type, 0);
  if(self) {
    self->t_min = ~0;
    self->t_max = 0;
  }
  return (PyObject *)self;
}

static void
cvs2svn_timerange_dealloc(cvs2svn_timerange *self)
{
  self->ob_type->tp_free((PyObject *)self);
}

static PyObject *
cvs2svn_timerange_add(cvs2svn_timerange *self, PyObject *timestamp)
{
  unsigned long val = -1;
  if (PyInt_CheckExact(timestamp)) {
    long i_long = PyInt_AS_LONG(timestamp);
    if(i_long < 0) {
       PyErr_SetString(PyExc_ValueError,
         "negative timestamp added");
       return NULL;
    }
    val = i_long;
  } else if(PyLong_Check(timestamp)) {
    val = PyLong_AsUnsignedLong(timestamp);
    // yes a goofy check, but has to be done since this is how they
    // return
    if((long)-1 == val && PyErr_Occurred())
       return NULL;
  } else {
    PyErr_SetString(PyExc_ValueError,
        "timestamp was neither an int or a long");
  }
  if(val < self->t_min)
    self->t_min = val;
  if(val > self->t_max)
    self->t_max = val;
  Py_RETURN_NONE;
}

static PyObject *
cvs2svn_timerange_richcompare(cvs2svn_timerange *self,
  cvs2svn_timerange *other, int op)
{
  if(self->ob_type != other->ob_type) {
    PyErr_SetString(PyExc_TypeError,
      "can only compare timerange to timerange");
    return NULL;
  }

  int result = self->t_max == other->t_max;
  switch(op)
  {
    case Py_EQ:
      if(result)
        result = (self->t_min == other->t_min);
      break;
    case Py_NE:
      if(result)
        result = (self->t_min != other->t_min);
      else
        result = 1;
      break;
    case Py_LT:
      if(result) {
        result = self->t_min < other->t_min;
      } else { 
        result = self->t_max < other->t_max;
      }
      break;
    case Py_LE:
      if(result) {
        result = self->t_min <= other->t_min;
      } else {
        result = self->t_max <= other->t_max;
      }
      break;
    case Py_GT:
      if(result) {
        result = self->t_min > other->t_min;
      } else { 
        result = self->t_max > other->t_max;
      }
      break;
    case Py_GE:
      if(result) {
        result = self->t_min >= other->t_min;
      } else {
        result = self->t_max >= other->t_max;
      }
      break;
    default:
      PyErr_SetString(PyExc_SystemError,
        "got nonstandard richcmp op");
        return NULL;
  }
  if(result) {
    Py_RETURN_TRUE;
  }
  Py_RETURN_FALSE;
}

int
cvs2svn_timerange_cmp(cvs2svn_timerange *self,
  cvs2svn_timerange *other)
{
  if(self->ob_type != other->ob_type) {
    PyErr_SetString(PyExc_TypeError,
      "can only compare timerange to timerange");
    return -2;
  }
  signed long long ret = self->t_max - other->t_max;
  if(0 != ret) {
    ret = self->t_min - other->t_min;
  }
  if(ret > 0) {
    ret = 1;
  } else if(ret < 0) {
    ret = -1;
  }
  return ret;
}

static PyMemberDef cvs2svn_timerange_members[] = {
  {"t_min", T_ULONG, offsetof(cvs2svn_timerange, t_min), 0, "t_min"},
  {"t_max", T_ULONG, offsetof(cvs2svn_timerange, t_max), 0, "t_max"},
  {NULL}
};

static PyMethodDef cvs2svn_timerange_methods[] = {
  {"add", (PyCFunction)cvs2svn_timerange_add, METH_O},
  {NULL}
};

static PyTypeObject cvs2svn_TimeRange_Type = {
  PyObject_HEAD_INIT(NULL)
  0,                                               /* ob_size*/
  "cvs2svn_lib.time_range.TimeRange",
                                                   /* tp_name*/
  sizeof(cvs2svn_timerange),                       /* tp_basicsize*/
  0,                                               /* tp_itemsize*/
  (destructor)cvs2svn_timerange_dealloc,           /* tp_dealloc*/
  0,                                               /* tp_print*/
  0,                                               /* tp_getattr*/
  0,                                               /* tp_setattr*/
  (cmpfunc)cvs2svn_timerange_cmp,                  /* tp_compare*/
  0,                                               /* tp_repr*/
  0,                                               /* tp_as_number*/
  0,                                               /* tp_as_sequence*/
  0,                                               /* tp_as_mapping*/
  0,                                               /* tp_hash */
  (ternaryfunc)0,                                  /* tp_call*/
  (reprfunc)0,                                     /* tp_str*/
  0,                                               /* tp_getattro*/
  0,                                               /* tp_setattro*/
  0,                                               /* tp_as_buffer*/
  Py_TPFLAGS_BASETYPE|Py_TPFLAGS_DEFAULT,          /* tp_flags*/
  0,                                               /* tp_doc */
  (traverseproc)0,                                 /* tp_traverse */
  (inquiry)0,                                      /* tp_clear */
  (richcmpfunc)cvs2svn_timerange_richcompare,      /* tp_richcompare */
  0,                                               /* tp_weaklistoffset */
  (getiterfunc)0,                                  /* tp_iter */
  (iternextfunc)0,                                 /* tp_iternext */
  cvs2svn_timerange_methods,                       /* tp_methods */
  cvs2svn_timerange_members,                       /* tp_members */
  0,                                               /* tp_getset */
  0,                                               /* tp_base */
  0,                                               /* tp_dict */
  0,                                               /* tp_descr_get */
  0,                                               /* tp_descr_set */
  0,                                               /* tp_dictoffset */
  (initproc)0,                                     /* tp_init */
  0,                                               /* tp_alloc */
  cvs2svn_timerange_new,                           /* tp_new */
};

PyDoc_STRVAR(
    cvs2svn_timerange_documentation,
    "cpython TimeRange for speed");

PyMODINIT_FUNC
init_time_range()
{
  PyObject *m = Py_InitModule3("_time_range", NULL,
    cvs2svn_timerange_documentation);
  if (!m)
    return;

  if (PyType_Ready(&cvs2svn_TimeRange_Type) < 0)
    return;

  Py_INCREF(&cvs2svn_TimeRange_Type);
  if (PyModule_AddObject(
    m, "TimeRange", (PyObject *)&cvs2svn_TimeRange_Type) == -1)
  return;
}
